// $Id: djinniprocess-run.c,v 1.5 2004/01/09 01:01:55 ensc Exp $    --*- c -*--

// Copyright (C) 2003 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "djinniprocess.h"
#include "djinniprocess-internal.h"
#include "wrappers.h"
#include "wish.h"
#include "parameters.h"
#include "command.h"
#include "util-io.h"
#include "lib/fmt.h"

#include <sys/select.h>
#include <sys/param.h>
#include <assert.h>
#include <sys/un.h>
#include <sys/wait.h>

static struct Credentials const		ROOT_CREDS = { 0,0,0 };
static int const			ON = 1;

static void
setFds(struct DjinniProcess *self, fd_set *read_fds)
{
  FD_ZERO(read_fds);

  FD_SET(self->fd_cmd_master,        read_fds);
  FD_SET(self->fd_cmd_vserver,       read_fds);

  if (self->fd_heartbeat_vserver!=-1)
    FD_SET(self->fd_heartbeat_vserver, read_fds);

  if (self->fd_parent!=-1)
    FD_SET(self->fd_parent, read_fds);
}

// Wait until we get a signal. Possible signal-sources are:
// - fd_cmd_master        ...  can be 'quit' only -> aborting later
// - fd_parent            ...  parent died        -> aborting later
// - fd_heartbeat_vserver ...  djinni-rub process has been attached -> continue
//
// In latter case, we are sending the '.' synchronization signal back to the
// slave and setting 'fd_heartbeat_vserver' to the new heartbeat socket
static void
waitOnHeartBeat(struct DjinniProcess *self)
{
  assert(self->fd_heartbeat_vserver!=-1);

  int			fd;
  struct sockaddr_un	addr;
  socklen_t		len = sizeof addr;
  fd_set		fds;
  int			max_fd = MAX(self->fd_heartbeat_vserver,
				     self->fd_cmd_vserver);

  FD_ZERO(&fds);
  FD_SET(self->fd_cmd_master,        &fds);
  FD_SET(self->fd_heartbeat_vserver, &fds);
  
  if (self->fd_parent!=-1) {
    FD_SET(self->fd_parent, &fds);
    max_fd = MAX(max_fd, self->fd_parent);
  }
  
  Elisten(self->fd_heartbeat_vserver, 5);
  while (Eselect(max_fd+1, &fds, 0,0, 0)==0) {}

  if (FD_ISSET(self->fd_heartbeat_vserver, &fds)) {
    fd = Eaccept(self->fd_heartbeat_vserver, &addr, &len);
    setCloseOnExec(fd);
    Elisten(self->fd_cmd_vserver, 5);
  }
  else
    fd = -1;

  Eclose(self->fd_heartbeat_vserver);
  self->fd_heartbeat_vserver = fd;

  if (fd!=-1) {
    char		c = '.';
    Ewrite(fd, &c, sizeof c);
  }
}

static bool
setCredentials(struct DjinniProcess *self, struct Credentials const *creds)
{
  assert(!self->credentials.f);
  
  self->credentials.f = true;
  self->credentials.v = *creds;

  DjinniProcess_chdirState(self);
  int		fd = Eopen(".ctx", O_WRONLY|O_EXCL|O_CREAT, 0644);
  char		buf[sizeof(xid_t)*4 + 2];
  size_t	len = fmt_long(buf, creds->ctx);

  buf[len] = '\0';

  Ewrite(fd, buf, len);
  Ewrite(fd, "\n", 1);
  Eclose(fd);

  Esetenv("DJINNI_CTX", buf, 1);
  return true;
}

static void
handleRequest(struct DjinniProcess *self, int sock_fd, bool need_auth)
{
  struct sockaddr_un	addr;
  socklen_t		len = sizeof addr;
  int			fd = Eaccept(sock_fd, &addr, &len);
  struct Wish		wish;
  struct Command	command;
  struct Credentials const *creds;

  if (!need_auth)               creds = &ROOT_CREDS;
  else if (self->credentials.f) creds = &self->credentials.v;
  else                          creds = 0;

  setCloseOnExec(fd);
  Esetsockopt(fd, SOL_SOCKET, SO_PASSCRED, &ON, sizeof(ON));

  if (Command_init(&command, self, fd) &&
      Wish_init(&wish, command.fd, creds) &&
      Wish_recv(&wish, &command.params) &&
      (creds=Wish_getCredentials(&wish))!=0 &&
      
      ((!need_auth ||
	self->credentials.f ||
	setCredentials(self, creds)))) {
    
    Command_execute(&command);
  }
  else {
    // TODO: logging
  }
  
  Wish_delete(&wish);
  Command_delete(&command);

  // command took control over 'fd' and closes it
}

static void
zombieKillHandler(int UNUSED sig)
{
  while (wait4(-1,0,WNOHANG,0)>0) {}
}

int
DjinniProcess_run(struct DjinniProcess *self)
{
  assert(self->fd_cmd_vserver!=-1);
  assert(self->fd_cmd_master!=-1);

  int			max_fd    = MAX(self->fd_cmd_master,
					self->fd_cmd_vserver);

  Elisten(self->fd_cmd_master, 1);
  if (self->fd_heartbeat_vserver!=-1)
    waitOnHeartBeat(self);
  else
    Elisten(self->fd_cmd_vserver, 5);

  if (self->fd_heartbeat_vserver!=-1)
    max_fd = MAX(max_fd, self->fd_heartbeat_vserver);
  if (self->fd_parent!=-1)
    max_fd = MAX(max_fd, self->fd_parent);

  ++max_fd;

  while (self->is_alive) {
    fd_set		read_fds;
    int			cnt;

    setFds(self, &read_fds);

    // TODO: set signal-handler
    zombieKillHandler(0);
    cnt = Eselect(max_fd, &read_fds, 0, 0, 0);

    if ((self->fd_parent!=-1 &&
	 FD_ISSET(self->fd_parent, &read_fds)) ||
	(self->fd_heartbeat_vserver!=-1 &&
	 FD_ISSET(self->fd_heartbeat_vserver, &read_fds))) break;

    if (FD_ISSET(self->fd_cmd_master,  &read_fds))
      handleRequest(self, self->fd_cmd_master,  false);
    if (FD_ISSET(self->fd_cmd_vserver, &read_fds))
      handleRequest(self, self->fd_cmd_vserver, true);
  }

  return 0;
}
