// $Id: syntaxinfo.c,v 1.1 2003/11/04 22:32:40 ensc Exp $    --*- c -*--

// Copyright (C) 2003 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "syntaxinfo.h"
#include "util.h"

#include <limits.h>
#include <assert.h>
#include <string.h>

static void
freeLimited(struct SyntaxInfo *info, size_t cnt)
{
  size_t		i;
  for (i=0; i<cnt; ++i) free(const_cast(char *)(info->info.set.v[i]));
  free(info->info.set.v);
}

// parses '' or '[(<val>,)*<val>?]'
inline static bool
parseSet(struct SyntaxInfo *info, char const **ptr)
{
  char const	*c  = *ptr;
  size_t	cnt = 0;
  size_t	i;
  bool		is_escaped = false;
  bool		is_eol     = false;

  if (**ptr=='\n' || **ptr=='\0') {
    info->info.set.l = 0;
    info->info.set.v = 0;

    return true;
  }
  
  if (**ptr!='[') return false;
  
  while (*c!='\0' && !is_eol) {
    if (is_escaped) is_escaped = true;
    else switch (*c) {
      case '\\'	:  is_escaped = true; break;
      case ','	:  ++cnt; break;
      case ']'	:  ++cnt; /*@fallthrough@*/
      case '\n'	:  is_eol = true; break;
      default	:  break;
    }

    ++c;
  }

  if (is_escaped && *c=='\0') return false;

  info->info.set.l = cnt;
  info->info.set.v = malloc(sizeof(char const *) * cnt);

  if (info->info.set.v==0)    return false;

  is_escaped = false;
  is_eol     = false;
  c          = *ptr+1;
  i          = 0;
  
  while (!is_eol) {
    if (is_escaped) is_escaped=false;
    else switch (**ptr) {
      case '\\'	:  is_escaped = true; break;
      case ']'	:  is_eol     = true; /*@fallthrough@*/
      case ','	:
	assert(i<cnt);
	info->info.set.v[i] = malloc(sizeof(char) * (*ptr-c) + 1);
	if (info->info.set.v[i]==0) {
	  freeLimited(info, i);
	  return false;
	}
	memcpy(info->info.set.v[i], c, *ptr-c);
	info->info.set.v[i][*ptr-c] = '\0';
	++i;
	c = *ptr+1;
	break;
      default	:  break;
    }
    ++*ptr;
  }

  assert((*ptr)[-1]==']');
  return true;
}

bool
parseIntRange(struct SyntaxInfo *info, char const **ptr)
{
  info->info.i.min = INT_MIN;
  info->info.i.max = INT_MAX;

  if (**ptr=='\n') return true;
  if (**ptr==',') ++*ptr;
  else {
    char	*errptr;
    info->info.i.min = strtol(*ptr, &errptr, 0);
    if (*errptr=='\n' || *errptr=='\0') {
      info->info.i.max = info->info.i.min;
      *ptr = errptr;
      return true;
    }
    if (*errptr!=',') return false;
    *ptr = errptr+1;
  }
  
  if (**ptr!='\0' && **ptr!='\n') {
    char	*errptr;
    info->info.i.max = strtol(*ptr, &errptr, 0);
    if (*errptr!='\n' && *errptr!='\0') return false;
    
    *ptr = errptr;
  }
  return true;
}

bool
parseUIntRange(struct SyntaxInfo *info, char const **ptr)
{
  info->info.i.min = 0;
  info->info.i.max = UINT_MAX;

  if (**ptr=='\n') return true;
  if (**ptr==',') ++*ptr;
  else {
    char	*errptr;
    info->info.i.min = strtol(*ptr, &errptr, 0);
    if (*errptr=='\n' || *errptr=='\0') {
      info->info.i.max = info->info.i.min;
      *ptr = errptr;
      return true;
    }
    if (*errptr!=',') return false;
    *ptr = errptr+1;
  }
  if (**ptr!='\0' && **ptr!='\n') {
    char	*errptr;
    info->info.i.max = strtol(*ptr, &errptr, 0);
    if (*errptr!='\n' && *errptr!='\0') return false;
    
    *ptr = errptr;
  }
  return true;
}

inline static void
moveToNextInfo(char const **ptr)
{
  bool		is_comment = false;

  while (**ptr!='\0') {
    if (is_comment) {
      if (**ptr=='\0') is_comment = false;
    }
    else switch (**ptr) {
      case '#'	:  is_comment = true; break;
      case ' '	:
      case '\n'	:
      case '\t'	:  break;
      default	:  return;
    }

    ++*ptr;
  }
}

bool
SyntaxInfo_init(struct SyntaxInfo *info, char const **ptr)
{
  moveToNextInfo(ptr);
  
  switch (**ptr) {
    case '\0'	:  return false;
    case 'v'	:  info->type = synVSERVER; ++*ptr; break;
    case 'i'	:  info->type = synINT;     ++*ptr; break;
    case 'u'	:  info->type = synUINT;    ++*ptr; break;
    case 'p'	:  info->type = synPATH;    ++*ptr; break;
    case 's'	:  info->type = synSTRING;  ++*ptr; break;
    case '['	:  info->type = synSET;             break;
    default	:  return false;
  };

  switch (info->type) {
    case synVSERVER	:
    case synSET		:  return parseSet(info, ptr);
    case synINT		:  return parseIntRange(info, ptr);
    case synUINT	:  return parseUIntRange(info, ptr);
    case synPATH	:
    case synSTRING	:  return parseUIntRange(info, ptr);
    default		:  assert(false); return false;
  }

  switch (**ptr) {
    case '\n'		:  ++*ptr; /*@fallthrough@*/
    case '\0'		:  return true;
    default		:  return false;
  }
}


void
SyntaxInfo_destroy(struct SyntaxInfo *info)
{
  switch (info->type) {
    case synVSERVER	:
    case synSET		:  freeLimited(info, info->info.set.l); break;
    default		:  break;
  }
}
